from google.cloud.firestore import Query as FSQuery
from firestore.db import MalformedQueryError, ListProperty


class Query(object):
    """
    A  query object is returned when you call :class:`~.mantle.firestore.db.Entity`.query().
    You can iterate over the query to get the results of your query one by one. Each item is an instance of a
    :class:`Model`
    """

    def __init__(self, entity, offset: int, limit: int):
        """Initialize a query

        Args:
            entity: The entity you want to query
            offset: The position to begin the query results
            limit: Maximum number of results to return
        """
        print(entity)
        from firestore.entity import __get_client__
        self.__query = __get_client__().collection(entity.__name__)
        if offset:
            self.__query = self.__query.offset(offset)
        if limit:
            self.__query = self.__query.limit(limit)
        self.__fetched = False
        self.__entity = entity
        self.__array_contains_queries = 0
        self.__range_filter_queries = {}

    def __validate_value(self, property_name, value):
        property = getattr(self.__entity, property_name)
        return property.__get_base_value__(value)

    def __add_range_filter(self, property):
        self.__range_filter_queries[property] = True

        # Range filter queries are only allowed on a single property at any given time
        if len(self.__range_filter_queries.keys()) > 1:
            raise MalformedQueryError("Range filter queries i.e (<), (>), (<=) and (>=) "
                                      "can only be performed on a single property in a query")

    def equal(self, property, value):
        """
        A query condition where property == value

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__query = self.__query.where(property, "==", self.__validate_value(property, value))
        return self

    def greater_than(self, property, value):
        """
        A query condition where property > value

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(property)
        self.__query = self.__query.where(property, ">", self.__validate_value(property, value))

    def less_than(self, property, value):
        """
        A query condition where property < value

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(property)
        self.__query = self.__query.where(property, "<", self.__validate_value(property, value))

    def greater_than_or_equal(self, property, value):
        """
        A query condition where property >= value

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(property)
        self.__query = self.__query.where(property, ">=", self.__validate_value(property, value))

    def less_than_or_equal(self, property, value):
        """
        A query condition where property <= value

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added
        """
        self.__add_range_filter(property)
        self.__query = self.__query.where(property, "<=", self.__validate_value(property, value))

    def contains(self, property, value):
        """
        A query condition where `value in property`

        Args:
             property (str): The name of a property to compare
             value (Any): The value to compare from the property

        Returns:
            Query: A query object with this condition added

        Raises:
            MalformedQueryError: If the property specified is not a ListField, or
               the query has more than one contains condition
        """
        model_property = getattr(self.__entity, property)

        # Don't do a contains condition in an invalid property
        if not isinstance(model_property, ListProperty):
            raise MalformedQueryError("Invalid property %s, query property for contains must be a list" % property)

        # Make sure there's only on `array_contains` condition
        self.__array_contains_queries += 1
        if self.__array_contains_queries > 1:
            raise MalformedQueryError("Only one `contains` clause is allowed per query")
        self.__query = self.__query.where(property, "array_contains", value)
        return self

    def order_by(self, property, direction="ASC"):
        """
        Set an order for the query, accepts

        Args:
            property (str): The property name to order by
            direction (str: "ASC" or "DESC"), optional:

        Returns:
             Query: A query object with order applied
        """
        if direction is not "ASC" and direction is not "DESC":
            raise MalformedQueryError("order_by direction can only be ASC, or DESC")
        direction = FSQuery.ASCENDING if direction is "ASC" else FSQuery.DESCENDING
        self.__query = self.__query.order_by(property, direction=direction)
        return self

    def __fetch(self):
        self.__docs = self.__query.get()
        self.__fetched = True

    def __iter__(self):
        return self

    def fetch(self):
        """
        Get the results of the query as a list

        Returns:
            list (Model): A list of models for the found results
        """
        return [entity for entity in self]

    def __next__(self):
        # Fetch data from db if not already done
        if not self.__fetched:
            self.__fetch()
        doc = self.__docs.__next__()
        user_data = self.__entity.__get_user_data__(doc.to_dict())
        return self.__entity(id=doc.id, **user_data)



"""
This is a collection of tools used by mantle Database packages, the include property types and common errors
"""

import json
import pickle
from datetime import datetime, date
from google.cloud.firestore import SERVER_TIMESTAMP


class Property(object):
    """
    A class describing a typed, persisted attribute of a database entity
    """
    def __init__(self, default=None, required=False):
        """
        Args:
            default: The default value of the property
            required: Enforce the property value to be provided
        """
        if type(self) is Property:
            raise Exception("You must extend Property")
        self.default = default
        self.required = required
        self.name = None

    def __get_user_value__(self, base_value):
        """
        Convert value from database to a value usable by the user
        Args:
            base_value: The current value from db

        Returns:
            user_value expected to be of the specified type
        """
        raise NotImplementedError

    def __get_base_value__(self, user_value):
        """
        Convert value to database acceptable format
        Args
        :param user_value: Current user_value
        :return: base_value
        """
        raise NotImplementedError

    def __type_check__(self, user_value, data_types):
        """
        Check whether this value has the right data type
        Args:
            user_value: The user_value you want to confirm
            data_types: Type/Types to check against

        Returns:
            Any
        """
        if self.required and self.default is None and user_value is None:
            raise InvalidValueError(self, user_value)
        # Assign a default value if None is provided
        if user_value is None:
            user_value = self.default

        if not isinstance(user_value, data_types) and user_value is not None:
            raise InvalidValueError(self, user_value)
        return user_value


class TextProperty(Property):
    """An Property whose value is a text string of unlimited length.
    I'ts not advisable to index this property
    """
    def __get_base_value__(self, user_value):
        return self.__type_check__(user_value, str)

    def __get_user_value__(self, base_value):
        return base_value


class StringProperty(Property):
    """
    An indexed Property whose value is a text string of limited length.

    Args:
        default: Default value for this property
        length (int): The maximum length of this property
        required (bool): Enforce whether this value can be empty
    """
    def __init__(self, default=None, length=255, required=False):
        super(StringProperty, self).__init__(default=default, required=required)
        self.length = length

    def __get_base_value__(self, user_value):
        user_value = self.__type_check__(user_value, str)
        if user_value is not None and len(user_value) > self.length:
            raise InvalidValueError(self, user_value)
        return user_value

    def __get_user_value__(self, base_value):
        return base_value


class IntegerProperty(Property):
    """A Property whose value is a Python int or long"""
    def __get_base_value__(self, user_value):
        return self.__type_check__(user_value, int)

    def __get_user_value__(self, base_value):
        return base_value


class FloatingPointNumberProperty(Property):
    """A Property whose value is a Python float.

    Note: int and long are also allowed.
    """
    def __get_base_value__(self, user_value: float):
        user_value = self.__type_check__(user_value, (int, float))
        return user_value

    def __get_user_value__(self, base_value):
        return base_value


class BlobProperty(Property):
    """A Property whose value is a byte string. It may be compressed."""
    def __get_base_value__(self, user_value):
        return self.__type_check__(user_value, (bytes, bytearray))

    def __get_user_value__(self, base_value):
        return base_value


class ListProperty(Property):
    """A List property"""
    def __init__(self, property_type: Property):
        super(ListProperty, self).__init__(default=[])
        self.property_type = property_type

    def __get_base_value__(self, user_value: list):
        user_value = self.__type_check__(user_value, (list))
        user_value = [self.property_type.__get_base_value__(value) for value in user_value]
        return user_value

    def __get_user_value__(self, base_value):
        return base_value


class ReferenceProperty(Property):
    def __init__(self, entity, required=False):
        from firestore import Entity
        if not issubclass(entity, Entity):
            raise ReferencePropertyError("A reference property must reference another model")
        super(ReferenceProperty, self).__init__(required=required)
        self.entity = entity

    def __get_base_value__(self, user_value):
        user_value = self.__type_check__(user_value, (self.entity))
        if not user_value:
            return user_value
        if not user_value.id:
            raise ReferencePropertyError("A reference must be put first before it can be referenced")
        return user_value.__document__()

    def __get_user_value__(self, base_value):
        if not base_value:
            return None
        user_data = self.entity.__get_user_data__(base_value.get().to_dict())
        return self.entity(id=base_value.id, **user_data)


class JsonProperty(Property):
    """A property whose value is any Json-encodable Python object.
    """
    def __init__(self, required=False):
        super(JsonProperty, self).__init__(required=required, default={})

    def __get_base_value__(self, user_value):
        if isinstance(user_value, str):
            try:
                user_value = json.loads(user_value)
            except json.JSONDecodeError:
                raise InvalidPropertyError(self, "Dict property must be valid JSON")
        return self.__type_check__(user_value, dict)

    def __get_user_value__(self, base_value):
        return base_value


class BooleanProperty(Property):
    """A Property whose value is a Python bool."""
    def __get_base_value__(self, user_value):
        return self.__type_check__(user_value, bool)

    def __get_user_value__(self, base_value):
        return base_value


class DateTimeProperty(Property):
    """A Property whose value is a datetime object.

    Note: auto_now_add can be overridden by setting the value before writing the entity.
    Args:
        default (datetime)
        required (bool): Enforce that this property can't be submitted when empty
        auto_now (bool): Set to the current time every time the model is updated
        auto_add_now (bool): Set to the current time when a record is created
    """
    def __init__(self, default=None, required=False, auto_now=False, auto_add_now=False):
        if not default and auto_add_now:
            default = SERVER_TIMESTAMP
        super(DateTimeProperty, self).__init__(default=default, required=required)
        self.auto_now = auto_now

    def __get_base_value__(self, user_value):
        # Return server timestamp as the value
        if user_value == SERVER_TIMESTAMP or self.auto_now:
            return SERVER_TIMESTAMP
        if user_value is None and self.default == SERVER_TIMESTAMP:
            return SERVER_TIMESTAMP
        return self.__type_check__(user_value, datetime)

    def __get_user_value__(self, base_value):
        return base_value


class DateProperty(Property):
    """A Property whose value is a date object.

    Args:
        default (datetime): The default value for this property
        required (bool): Enforce that this property can't be submitted when empty
        auto_add_now (bool): Set to the current date when a record is created
    """
    def __init__(self, default=None, required=False, auto_add_now=False):
        if not default and auto_add_now:
            default = SERVER_TIMESTAMP
        super(DateProperty, self).__init__(default=default, required=required)

    def __get_base_value__(self, user_value):
        # Return server timestamp as the value
        if user_value is None and self.default == SERVER_TIMESTAMP:
            return SERVER_TIMESTAMP
        if isinstance(user_value, datetime):
            return user_value.date()
        return self.__type_check__(user_value, date)

    def __get_user_value__(self, base_value):
        if isinstance(base_value, datetime):
            return base_value.date()
        return base_value


class PickledProperty(Property):
    """A Property whose value is any picklable Python object."""
    def __int__(self, required=False):
        super(PickledProperty, self).__init__(default=None, required=required)

    def __get_base_value__(self, user_value):
        return pickle.dumps(user_value)

    def __get_user_value__(self, base_value):
        return pickle.loads(base_value)


class InvalidValueError(ValueError):
    """Raised if the value of a property does not fit the property type"""

    def __init__(self, property, value):
        self.property = property
        self.value = value

    def __str__(self):
        return "%s is not a valid value for property %s of type %s" % \
               (self.value, self.property.name, type(self.property).__name__)


class MalformedQueryError(Exception):
    """Raised when the rules of a query are broken"""

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class InvalidPropertyError(Exception):
    """Raised if a non-existent property is provided during the creation of a model"""

    def __init__(self, prop_name, model_name):
        self.prop_name = prop_name
        self.model_name = model_name

    def __str__(self):
        return "%s not found in model %s" % (self.prop_name, self.model_name)


class ReferencePropertyError(Exception):
    """Raised when a reference property point's to a location the model can't resolve"""
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message



import setuptools


install_requires = ["google-cloud-firestore"]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="firestore-entities",
    version="0.0.9",
    author="Billcountry Mwaniki",
    author_email="me@billcountry.tech",
    description="Implementation of entities concept on top of Google cloud firestore",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Billcountry/firestore-entities",
    packages=['firestore'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires
)



from google.cloud.firestore import Client
from firestore.db import InvalidPropertyError, Property
from firestore.query import Query


def __get_client__():
    # There must be a better way to do this, right?
    import builtins
    if not hasattr(builtins, "__firestore_client__"):
        setattr(builtins, "__firestore_client__", Client())
    return builtins.__firestore_client__


class Entity(object):
    """Creates a firestore document under the collection [YourEntity]

    Args:
        **data (kwargs): Values for properties in the new record, e.g User(name="Bob")

    Attributes:
        id (str or int): Unique id identifying this record,
            if auto-generated, this is not available before `put()`
    """

    def __init__(self, **data):
        if type(self) is Entity:
            raise Exception("You must extend Model")
        self.__setup_properties()
        self.__model_name = type(self).__name__
        client = __get_client__()
        self.__collection = client.collection(self.__collection_path())
        self.id = None
        if "id" in data:
            self.id = data.pop("id")
        for key, value in data.items():
            if key in self.__properties:
                setattr(self, key, value)
            else:
                raise InvalidPropertyError(key, self.__model_name)

    @classmethod
    def __collection_path(cls):
        return cls.__name__

    def __document__(self):
        if not self.id:
            return None
        # Get's the absolute path: `projects/{project_id}/databases/{database_id}/documents/{document_path}
        return self.__collection.document(self.id)

    def __str__(self):
        return "<Model %s>" % self.__model_name

    def __setup_properties(self):
        # Get defined properties, equate them to their defaults
        self.__properties = dict()
        for attribute in dir(self):
            if attribute.startswith("_"):
                continue
            value = getattr(self, attribute)
            if isinstance(value, Property):
                value.name = attribute
                self.__properties[attribute] = value
                setattr(self, attribute, value.default)

    def __prepare(self):
        # Find current property values and validate them
        values = dict()
        for key, prop in self.__properties.items():
            value = getattr(self, key)
            values[key] = prop.__get_base_value__(value)
        return values

    def put(self):
        """
        Save the models data to Firestore

        Raises:
            InvalidValueError: Raised if the value of a property is invalid, e.g. A required property that's None
        """
        data = self.__prepare()
        if self.id:
            self.__collection.document(self.id).set(data)
            return
        _time, new_ref = self.__collection.add(data)
        self.id = new_ref.id

    def delete(self):
        """Delete the document connected to this model from firestore"""
        if self.id:
            self.__collection.document(self.id).delete()

    @classmethod
    def __get_user_data__(cls, base_data):
        user_data = dict()
        for name in dir(cls):
            prop = getattr(cls, name)
            if not isinstance(prop, Property):
                continue
            if name in base_data:
                user_data[name] = prop.__get_user_value__(base_data.get(name))
            else:
                user_data[name] = prop.default
        return user_data

    @classmethod
    def get(cls, _id):
        """
        Get a model with the given id

        Args:
            _id (str or int): A key or id of the model record, when a list is provided, `get` returns a list
                models

        Returns:
            Entity: An instance of the firestore entity calling get
            None: If the id provided doesn't exist
        """
        document = __get_client__().collection(cls.__collection_path()).document(_id)
        data = document.get()
        if not data.exists:
            return None
        user_data = cls.__get_user_data__(data.to_dict())
        return cls(id=_id, **user_data)

    @classmethod
    def query(cls, offset=0, limit=0):
        """
        Create a query to this model

        Args:
            offset (int): The position in the database where the results begin
            limit (int): Maximum number of records to return

        Returns:
            An iterable query object
        """
        return Query(cls, offset, limit)
